@@ -27,12 +27,19 @@ class Pair:
 
 @dataclass
 class Line(ABC):
+    """
+    This class stores only a single reference to the column values and the column index. This is preferable to storing
+    the entire row as a dictionary, as it is more memory efficient and faster.
+    """
     _column_index: Dict[str, int]
     _values: List[Any]
 
     def __getitem__(self, item: str) -> Any:
         return process_value(self._values[self._column_index[item]])
 
+    def __setitem__(self, key: str, value: Any):
+        self._values[self._column_index[key]] = value
+
     @property
     def columns(self) -> List[str]:
         return list(self._column_index.keys())
@@ -44,62 +51,86 @@ class Line(ABC):
 
 @dataclass
 class SequenceLine(Line):
-    peak_area_light_label: str = 'PEAK_AREA_L'
-    peak_area_medium_label: str = 'PEAK_AREA_M'
-    peak_area_heavy_label: str = 'PEAK_AREA_H'
-    sequence_label: str = 'SEQUENCE'
-    charge_state_label: str = 'CS'
-    scan_number_label: str = 'SCAN'
-    filename_label: str = 'FILE_NAME'
-    xcorr_label: str = 'XCorr'
 
     @property
     def light_peak_area(self) -> float:
-        if self[self.peak_area_light_label] is None or math.isnan(self[self.peak_area_light_label]):
+        if self['PEAK_AREA_L'] is None or math.isnan(self['PEAK_AREA_L']):
             return 0.0
-        return float(self[self.peak_area_light_label])
+        return float(self['PEAK_AREA_L'])
+
+    @light_peak_area.setter
+    def light_peak_area(self, value: float):
+        self['PEAK_AREA_L'] = value
 
     @property
     def medium_peak_area(self) -> float:
-        if self[self.peak_area_medium_label] is None or math.isnan(self[self.peak_area_medium_label]):
+        if self['PEAK_AREA_M'] is None or math.isnan(self['PEAK_AREA_M']):
             return 0.0
-        return float(self[self.peak_area_medium_label])
+        return float(self['PEAK_AREA_M'])
+
+    @medium_peak_area.setter
+    def medium_peak_area(self, value: float):
+        self['PEAK_AREA_M'] = value
 
     @property
     def heavy_peak_area(self) -> float:
-        if self[self.peak_area_heavy_label] is None or math.isnan(self[self.peak_area_heavy_label]):
+        if self['PEAK_AREA_H'] is None or math.isnan(self['PEAK_AREA_H']):
             return 0.0
-        return float(self[self.peak_area_heavy_label])
+        return float(self['PEAK_AREA_H'])
+
+    @heavy_peak_area.setter
+    def heavy_peak_area(self, value: float):
+        self['PEAK_AREA_H'] = value
 
     @property
     def peptide_sequence(self) -> str | None:
-        if self[self.sequence_label] is None:
+        if self['SEQUENCE'] is None:
             return None
-        return str(self[self.sequence_label])
+        return str(self['SEQUENCE'])
+
+    @peptide_sequence.setter
+    def peptide_sequence(self, value: str):
+        self['SEQUENCE'] = value
 
     @property
     def peptide_charge(self) -> int | None:
-        if self[self.charge_state_label] is None:
+        if self['CS'] is None:
             return None
-        return int(self[self.charge_state_label])
+        return int(self['CS'])
+
+    @peptide_charge.setter
+    def peptide_charge(self, value: int):
+        self['CS'] = value
 
     @property
     def scannr(self) -> int | None:
-        if self[self.scan_number_label] is None:
+        if self['SCAN'] is None:
             return None
-        return int(self[self.scan_number_label])
+        return int(self['SCAN'])
+
+    @scannr.setter
+    def scannr(self, value: int):
+        self['SCAN'] = value
 
     @property
     def filename(self) -> str | None:
-        if self[self.filename_label] is None:
+        if self['FILE_NAME'] is None:
             return None
-        return str(self[self.filename_label])
+        return str(self['FILE_NAME'])
+
+    @filename.setter
+    def filename(self, value: str):
+        self['FILE_NAME'] = value
 
     @property
     def xcorr(self) -> float | None:
-        if self[self.xcorr_label] is None:
+        if self['XCorr'] is None:
             return None
-        return float(self[self.xcorr_label])
+        return float(self['XCorr'])
+
+    @xcorr.setter
+    def xcorr(self, value: float):
+        self['XCorr'] = value
 
 
 @dataclass
@@ -114,11 +145,13 @@ class SLine(SequenceLine):
 
 @dataclass
 class PLine(Line):
-    protein_label: str = 'LOCUS'
-
     @property
     def locus(self) -> str | None:
-        return str(self[self.protein_label])
+        return str(self['LOCUS'])
+
+    @locus.setter
+    def locus(self, value: str):
+        self['LOCUS'] = value
 
 
 @dataclass
@@ -191,6 +224,22 @@ class CPPResult:
 
         return loci
 
+    @property
+    def scannr(self) -> int:
+        return self.s_line.scannr
+
+    @property
+    def filename(self) -> str:
+        return self.s_line.filename
+
+    @filename.setter
+    def filename(self, value: str):
+        self.s_line.filename = value
+
+    @property
+    def charge(self) -> int:
+        return self.s_line.peptide_charge
+
 
 @dataclass
 class QuantResult:
@@ -202,6 +251,8 @@ class QuantResult:
 
     # Afterwards, since it is not known at the time of creation
     is_duplicate: bool = None
+    imputed_light: float = None
+    imputed_heavy: float = None
 
     @property
     def total_intensity(self) -> float:
@@ -221,11 +272,15 @@ class QuantResult:
 
     @property
     def charge(self) -> int:
-        return self.cpp_result.s_line.peptide_charge
+        return self.cpp_result.charge
 
     @property
     def filename(self) -> str:
-        return self.cpp_result.s_line.filename
+        return self.cpp_result.filename
+
+    @filename.setter
+    def filename(self, value: str):
+        self.cpp_result.filename = value
 
     @property
     def scannr(self) -> int:
@@ -260,8 +315,8 @@ class QuantResult:
         return self.light == 0.0 and self.heavy == 0.0
 
     @property
-    def duplicate_key(self) -> bool:
-        return (self.filename, self.scannr)
+    def duplicate_key(self) -> Tuple[str, int]:
+        return self.filename, self.scannr
 
     @property
     def loci(self) -> List[str]:
@@ -271,6 +326,10 @@ class QuantResult:
     def is_decoy(self) -> bool:
         return all('reverse' in protein.lower() for protein in self.loci)
 
+    @property
+    def is_contaminant(self) -> bool:
+        return any('contaminant' in protein for protein in self.loci)
+
     @property
     def is_double(self) -> bool:
         return self.cpp_result.is_double
@@ -281,22 +340,53 @@ class QuantResult:
 
     @property
     def light(self) -> float:
+
+        if self.imputed_light is not None:
+            return self.imputed_light
+
         return self.cpp_result.light[self.cpp_result_index]
 
     @property
     def heavy(self) -> float:
+        if self.imputed_heavy is not None:
+            return self.imputed_heavy
+
         return self.cpp_result.heavy[self.cpp_result_index]
 
     @cached_property
     def ratio(self) -> float:
-        return self.heavy / self.light
+
+        if self.heavy == 0.0 and self.light == 0.0:
+            return np.nan
+
+        if self.heavy == 0.0:
+            return np.inf
+
+        if self.light == 0.0:
+            return 0.0
+
+        return np.divide(self.light, self.heavy)
 
     @cached_property
     def log2_ratio(self) -> float:
+
+        if self.ratio == 0.0:
+            return -np.inf
+
+        if self.ratio == np.inf:
+            return np.inf
+
         return np.log2(self.ratio)
 
     @cached_property
     def log10_ratio(self) -> float:
+
+        if self.ratio == 0.0:
+            return -np.inf
+
+        if self.ratio == np.inf:
+            return np.inf
+
         return np.log10(self.ratio)
 
     @cached_property
@@ -371,12 +461,28 @@ class RatioResult:
         """
         return int(self._rollup[2])
 
-    def to_dict(self) -> Dict[str, Any]:
+    def to_dict(self, add_peptides: False, add_proteins: False) -> Dict[str, Any]:
         d = {c: self._get_quant_result_attribute(c) for c in self.grouping}
+        d2 = self.to_value_dict(add_peptides, add_proteins)
+
+        for k, v in d2.items():
+            d[k] = v
+
+        return d
+
+    def to_value_dict(self, add_peptides: False, add_proteins: False) -> Dict[str, Any]:
+        d = {}
         d['log2_ratio'] = self.log2_ratio
         d['log2_ratio_std'] = self.log2_ratio_std
         d['cnt'] = self.cnt
         d['type'] = self.ratio_type
+
+        if add_peptides:
+            d['peptide_site_str'] = ';'.join(self.peptide_site_strs)
+
+        if add_proteins:
+            d['protein_site_str'] = ';'.join(self.protein_site_strs)
+
         return d
 
     def _get_quant_result_attribute(self, attribute: str) -> Any:
@@ -401,6 +507,36 @@ class RatioResult:
     def non_group_key_labels(self) -> tuple[Any, ...]:
         return tuple([l for l, v in zip(self.grouping, self.grouping_vals) if l != 'group'])
 
+    @property
+    def ip2_sequences(self) -> List[str]:
+        if not self.is_valid:
+            return []
+        return list(set([qr.ip2_sequence for qr in self.quant_results]))
+
+    @property
+    def proforma_sequences(self) -> List[str]:
+        if not self.is_valid:
+            return []
+        return list(set([qr.proforma_sequence for qr in self.quant_results]))
+
+    @property
+    def unmodified_sequences(self) -> List[str]:
+        if not self.is_valid:
+            return []
+        return list(set([qr.unmodified_sequence for qr in self.quant_results]))
+
+    @property
+    def peptide_site_strs(self) -> List[str]:
+        if not self.is_valid:
+            return []
+        return list(set([qr.peptide_site_str for qr in self.quant_results]))
+
+    @property
+    def protein_site_strs(self) -> List[str]:
+        if not self.is_valid:
+            return []
+        return list(set([qr.protein_site_str for qr in self.quant_results]))
+
 
 @dataclass
 class CompareRatio:
@@ -503,17 +639,29 @@ class CompareRatio:
     @property
     def cnt(self) -> int:
         if not self.is_valid:
-            return np.nan
+            return 0
 
         return int(self.group1_ratio.cnt + self.group2_ratio.cnt)
 
-    def to_dict(self) -> Dict[str, Any]:
+    def to_dict(self, add_peptides: False, add_proteins: False) -> Dict[str, Any]:
 
         d1 = {}
         for l, v in zip(self.grouping, self.grouping_vals):
             d1[l] = v
 
-        d2 = {
+        d2 = self.to_value_dict(add_peptides, add_proteins)
+
+        if add_peptides:
+            d2['peptide_site_str'] = ';'.join(self.peptide_site_strs)
+
+        if add_proteins:
+            d2['protein_site_str'] = ';'.join(self.protein_site_strs)
+
+        d = {**d1, **d2}
+        return d
+
+    def to_value_dict(self, add_peptides: False, add_proteins: False) -> Dict[str, Any]:
+        d = {
             'group1': self.group1,
             'group2': self.group2,
             'group1_log2_ratio': self.group1_log2_ratio,
@@ -530,5 +678,70 @@ class CompareRatio:
             'qvalue': self.qvalue
         }
 
-        d = {**d1, **d2}
+        if add_peptides:
+            d['peptide_site_str'] = ';'.join(self.peptide_site_strs)
+
+        if add_proteins:
+            d['protein_site_str'] = ';'.join(self.protein_site_strs)
+
         return d
+
+    @property
+    def ip2_sequences(self) -> List[str]:
+        ip2_sequences = set()
+
+        if self.group1_ratio is not None:
+            ip2_sequences.update(self.group1_ratio.ip2_sequences)
+
+        if self.group2_ratio is not None:
+            ip2_sequences.update(self.group2_ratio.ip2_sequences)
+
+        return list(ip2_sequences)
+
+    @property
+    def proforma_sequences(self) -> List[str]:
+        proforma_sequences = set()
+
+        if self.group1_ratio is not None:
+            proforma_sequences.update(self.group1_ratio.proforma_sequences)
+
+        if self.group2_ratio is not None:
+            proforma_sequences.update(self.group2_ratio.proforma_sequences)
+
+        return list(proforma_sequences)
+
+
+    @property
+    def unmodified_sequences(self) -> List[str]:
+        unmodified_sequences = set()
+        if self.group1_ratio is not None:
+            unmodified_sequences.update(self.group1_ratio.unmodified_sequences)
+
+        if self.group2_ratio is not None:
+            unmodified_sequences.update(self.group2_ratio.unmodified_sequences)
+
+        return list(unmodified_sequences)
+
+    @property
+    def peptide_site_strs(self) -> List[str]:
+        peptide_site_strs = set()
+        if self.group1_ratio is not None:
+            peptide_site_strs.update(self.group1_ratio.peptide_site_strs)
+
+        if self.group2_ratio is not None:
+            peptide_site_strs.update(self.group2_ratio.peptide_site_strs)
+
+        return list(peptide_site_strs)
+
+    @property
+    def protein_site_strs(self) -> List[str]:
+        protein_site_strs = set()
+
+        if self.group1_ratio is not None:
+            protein_site_strs.update(self.group1_ratio.protein_site_strs)
+
+        if self.group2_ratio is not None:
+            protein_site_strs.update(self.group2_ratio.protein_site_strs)
+
+        return list(protein_site_strs)
+
